/*
 * Eprom Programmer
 *
 * Copyright 1996-2007 netFluid Technology Ltd
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.30 $  $Date: 2007/02/22 14:22:52 $
 * 
 * --------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sched.h>
#include <unistd.h>

#define _GNU_SOURCE
#include <getopt.h>

#include <sys/time.h>
#include <time.h>

#include "ceprog.h"
#include "device.h"
#include "delay.h"

const char WELCOME[] = "Welcome to the CEPROG Eprom Programmer (version "VERSION")\n(c) "VERSION_YEAR" netFluid Technology Ltd\n\n";

const char GOODBYE[]  = "Goodbye\n";
const char FINISHED[] = "Finished\n";

static struct option long_options[] = 
{
  {"help",         0, 0, 'h' },
  {"verbose",      0, 0, 'v' },
  {"read",         0, 0, 'r' },
  {"blank",        0, 0, 'b' },
  {"write",        0, 0, 'w' },
  {"quick",        0, 0, 'q' },
  {"verify",       0, 0, 'V' },
  {"show-devices", 0, 0, 'D' },
  {"realtime",     0, 0, 'R' },
  {"port",         1, 0, 'p' },
  {"device-id",    1, 0, 'i' },
  {"filename",     1, 0, 'f' },
  {0,0,0,0}
};

int verbose = 0;
int vindent = 0;

char *ModeDescription[] = { "Read", "Blank Check", "Standard Write", "Quick Write", "Verify" };


static unsigned char  control_lines  = 0x00;
static unsigned char  register_lines = 0x00;
static int            port_direction = 0;
static int            progPinCE      = 0;
static ProgrammerMode progMode       = MODE_READ;


static void program_off( Device *dtype, int device );
static void program_on ( Device *dtype, int device );

/* ------------------------------------------------------------------------- */

static void
usage()
{
    printf( "\
\n\
ceprog [-hvrbwqVR] -p parallel port device -i deviceID [-f file] [-D]\n\
 -h --help         Display this help\n\
 -v --verify       Verify EPROM against file\n\
 -r --read         Read EPROM, write to file if supplied\n\
 -b --blank        Read EPROM and confirm that it is blank\n\
 -w --write        Write file to EPROM using standard algorithm\n\
 -q --quick        Write file to EPROM using quick-write algorithm\n\
 -V --verbose      Switch on verbose mode (multiple -v to increase level)\n\
 -R --realtime     Enter realtime mode for precise timing\n\
 -D --show-devices List available devices and their IDs\n\
 -p --pport        Specify which parallel port device to use (eg /dev/parport0)\n\
 -f --filename     The name of the file to which data is read or written\n\
 -i --device-id    Specify the ID of the device in the programmer (see -D)\n\
\n" );


}

/* ------------------------------------------------------------------------- */

static void
prompt( char *msg )
{
    fflush(stdin);
    printf( "%s - press <Enter>\n", msg );
    fgetc(stdin);
}

/* ------------------------------------------------------------------------- */

static int
set_write_mode( int device )
{
    int mode = 0; /* Write Mode */

    if( port_direction == PORT_DIR_WRITE ) return 0;

    if (ioctl (device, PPDATADIR, &mode)) {
        perror ("PPDATADIR");
        return -1;
    }
    port_direction = PORT_DIR_WRITE;
    return 0;
}

/* ------------------------------------------------------------------------- */

static int
set_read_mode( int device )
{
    int mode = 1; /* Read Mode */

    if( port_direction == PORT_DIR_READ ) return 0;

    if (ioctl (device, PPDATADIR, &mode)) {
        perror ("PPDATADIR");
        return -1;
    }
    port_direction = PORT_DIR_READ;
    return 0;
}

/* ------------------------------------------------------------------------- */

static int
write_data( int device, unsigned char data )
{
    set_write_mode( device );

    if( verbose == 3 ) printf( "%*sWrite Data 0x%02x\n", vindent, "", data );

    if (ioctl (device, PPWDATA, &data)) {
        perror ("PPWDATA");
        return 1;
    }

    return 0;
}

/* ------------------------------------------------------------------------- */

static unsigned char
read_data( int device )
{
    unsigned char data = 0;

    set_read_mode( device );

    if (ioctl (device, PPRDATA, &data)) {
        perror ("PPWDATA");
        return 0;
    }
    if( verbose == 3 ) printf( "%*sRead byte from EPROM 0x%02x\n", vindent, "", data );

    return data;
}

/* ------------------------------------------------------------------------- */

static int
set_ctrl_line( int device, unsigned int pin )
{
    int ret = 0;
    unsigned char c = 0;

    c = pin & 0xff;
    if( verbose == 3 ) printf( "%*sSet PORT ctrl line %d\n", vindent, "", c );

    control_lines = (control_lines & ~c) | c;

    if( verbose == 4 ) printf( "ctrl = 0x%02x & 0x%02x | 0x%02x    -> 0x%02x\n", control_lines, ~c & 0xff, c, control_lines );

    c = control_lines ^ 0x0b; /* Toggle state of inverted lines */
    if (ioctl (device, PPWCONTROL, &c)) {
        perror ("PPWCONTROL");
        ret = -1;
    }
    return ret;
}

/* ------------------------------------------------------------------------- */

static int
reset_ctrl_line( int device, unsigned int pin )
{
    int ret = 0;
    unsigned char c = 0;

    c = pin & 0xff;
    if( verbose == 3 ) printf( "%*sReset PORT ctrl line %d\n", vindent, "", c );

    control_lines = (control_lines & ~c) | 0;
    if( verbose == 4 ) printf( "ctrl = 0x%02x & 0x%02x | 0x%02x    -> 0x%02x\n", control_lines, ~c & 0xff, 0, control_lines ); 

    c = control_lines ^ 0x0b; /* Toggle state of inverted lines */
    if (ioctl (device, PPWCONTROL, &c)) {
        perror ("PPWCONTROL");
        ret = -1;
    }
    return ret;
}

/* ------------------------------------------------------------------------- */

static int
write_register( int device, unsigned char data )
{
    unsigned char old_ctrl_state = control_lines;

    if( verbose == 3 ) printf( "%*s-> Pulse REG Latch pin Start\n", vindent,"");
    vindent += 2;

    /* Disable EPROM Ouput to clear the DataBus (if OE enabled) */
    if( !(control_lines & EPRG_EPROM_OE ) ) {
        if( verbose == 3 ) printf( "%*sDisable EPROM OE\n", vindent,"");
        set_ctrl_line( device, EPRG_EPROM_OE ); 
    }

    micro_sleep( 30 );

    write_data( device, data );
    write_data( device, data );
    write_data( device, data );

    /* Wait 7 ns  (t = Ds) */
    /* nano_sleep( 7 ); */

    /* Pulse REG_LATCH HIGH */
    set_ctrl_line( device, EPRG_REG_LATCH );
    set_ctrl_line( device, EPRG_REG_LATCH );
    set_ctrl_line( device, EPRG_REG_LATCH );

    /* Wait 5 ns ( t = tw ) */
    /* nano_sleep( 5 ); */

    reset_ctrl_line( device, EPRG_REG_LATCH );

    /* Wait 4 ns (t = th ) */
    /* nano_sleep( 4 ); */

    /* Put ourselves back into read mode */
    set_read_mode( device );

    /* Restore OE line to its previous active (low) state if it was in this
     * state on entry to write_register()
     */
    if( !(old_ctrl_state & EPRG_EPROM_OE ) ) {
      if( verbose == 3 ) printf( "%*sEnable EPROM OE\n", vindent,"");
      reset_ctrl_line( device, EPRG_EPROM_OE ); 
    }

    vindent -= 2;
    if( verbose == 3 ) printf( "%*s-> Pulse REG Latch pin Done\n", vindent,"");

    return 0;
}

/* ------------------------------------------------------------------------- */

static int
set_ctrl_reg( int device, unsigned int pin )
{
    unsigned char c = 0;

    if( pin & 0x0f ) {
        fprintf( stderr, "set_ctrl_reg Error: %d is not a valid control line\n", pin );
        return -1;
    }

    c = (pin & 0xff00) >> 8; /* Keep top 8 bits and shift into lower */

    if( verbose == 3 ) printf( "%*sSet REG  ctrl line %d\n", vindent, "", c );

    register_lines = (register_lines & ~c) | c;

    /* printf( "ctrl = 0x%02x & 0x%02x | 0x%02x    -> 0x%02x\n", control_lines, ~c & 0xff, c, control_lines ); */

    return write_register( device, register_lines );
}

/* ------------------------------------------------------------------------- */

static int
reset_ctrl_reg( int device, unsigned int pin )
{
    unsigned char c = 0;

    if( pin & 0x0f ) {
        fprintf( stderr, "reset_ctrl_reg Error: %d is not a valid control line\n", pin );
        return -1;
    }

    c = (pin & 0xff00) >> 8; /* Keep top 8 bits and shift into lower */

    if( verbose == 3 ) printf( "%*sReset REG  ctrl line %d\n", vindent, "", c );

    register_lines = (register_lines & ~c) | 0;

    /* printf( "ctrl = 0x%02x & 0x%02x | 0x%02x    -> 0x%02x\n", control_lines, ~c & 0xff, c, control_lines ); */

    return write_register( device, register_lines );
}

/* ------------------------------------------------------------------------- */

static int
set_control( int device, int line )
{
    int ret = 0;
    int ctrl;
    int latch;

    ctrl  = line & 0x000f;
    latch = line & 0xff00;
    /* printf( "set  : ctrl = 0x%02x  latch = 0x%04x\n", ctrl, latch ); */

    if( ctrl                ) ret = set_ctrl_line( device, ctrl  );
    if( (ret >= 0) && latch ) ret = set_ctrl_reg ( device, latch );

    return ret;
}

/* ------------------------------------------------------------------------- */

static int
reset_control( int device, int line )
{
    int ret = 0;
    int ctrl;
    int latch;

    ctrl  = line & 0x000f;
    latch = line & 0xff00;

    /* printf( "reset: ctrl = 0x%02x  latch = 0x%04x\n", ctrl, latch ); */

    if( ctrl                ) ret = reset_ctrl_line( device, ctrl  );
    if( (ret >= 0) && latch ) ret = reset_ctrl_reg ( device, latch );

    return ret;
}

/* ------------------------------------------------------------------------- */

static int
open_paraport( char *pport_name )
{
    int fd = -1;
    int mode = 0;

    /* Get permission to write directly to port (for delay) */
    if (ioperm(PPORT,2,1)<0)
    {
        perror("ioperm");
        return -1;
    }

    /* see http://people.redhat.com/twaugh/parport/html/x623.html */

    if( (fd = open( pport_name, O_RDWR )) < 0 )
    {
        perror( "Cannot open device\n" );
        return -1;
    }

    if( ioctl(fd, PPCLAIM) ) {
        perror ("PPCLAIM");
        close (fd);
        return -1;
    }

    mode = IEEE1284_MODE_ECP;
    if (ioctl (fd, PPSETMODE, &mode)) {
        perror ("PPSETMODE");
        close(fd);
        return -1;
    }
    return fd;
}

/* ------------------------------------------------------------------------- */

/* REset the address counters of the programmer
 */
static int
reset_counter( int device )
{
    if( verbose == 3 ) printf( "%*sReset programmer address counter...\n", vindent, "" );
    vindent += 2;

    /* Pulse EPRG_LATCH_CLOCK_RESET line low for a few usecs.  Note that 
     * this low pulse will have the length of the register_latch time
     * added to it....
     */
    set_control  ( device, EPRG_LATCH_CLOCK_RESET );
    set_control  ( device, EPRG_LATCH_CLOCK_RESET );
    set_control  ( device, EPRG_LATCH_CLOCK_RESET );

    /* Wait 7 ns  (t = tw) */
    /* nano_sleep( 7 ); */

    reset_control( device, EPRG_LATCH_CLOCK_RESET );
    reset_control( device, EPRG_LATCH_CLOCK_RESET );
    reset_control( device, EPRG_LATCH_CLOCK_RESET );
    /* nano_sleep( 7 ); */

    vindent -= 2;
    if( verbose == 3) printf( "%*sReset programmer address counter done\n", vindent, "" );

    return 0;
}

/* ------------------------------------------------------------------------- */

/* Advance the address counter
 */
static int
advance_counter( int device )
{
    if( verbose == 3 ) printf( "%*sAdvance programmer address counter...\n", vindent, "" );
    vindent += 2;

    /* Pulse EPRG_ADDR_CLOCK line high for a few nsecs.
     */
    set_control  ( device, EPRG_ADDR_CLOCK );
    set_control  ( device, EPRG_ADDR_CLOCK );

    /* Send the EPRG_ADDR_CLOCK line low. Counter clocks on the HIGH-to-LOW
     * clock transitions
     */
    reset_control( device, EPRG_ADDR_CLOCK );

    vindent -= 2;
    if( verbose == 3) printf( "%*sAdvance programmer address counter done\n", vindent, "" );

    return 0;
}

/* ------------------------------------------------------------------------- */
/* Enables EPROM by asserting active low E/CE pin
 */
static void
enable_eprom( int device )
{
    if( progPinCE ) {
        /* In this mode CE and Prog are the same pin. EPRG_LATCH_CE is 
         * overridden by the hardware by EPRG_EPROM_PROG so we use signal to
         * enable the eprom. However we cannot put this line * LOW if we're in
         * programming mode as this will send a 'write' pulse to the EPROM.
         */
        if (progMode == MODE_STD_WRITE || progMode == MODE_QUICK_WRITE) 
        {
            if( verbose ) fprintf( stderr, "\nAttempt to enable EPROM in WRITE MODE will send CE low and issue a PROG Pulse!\nenable_eprom() skipped....\n") ;
            return;
        }
        reset_control( device, EPRG_EPROM_PROG );
    } else {
        /* Prog and CE are different pins ... okay to drop CE line low */
        reset_control( device, EPRG_LATCH_CE );
    }

    micro_sleep(100);
}

/* ------------------------------------------------------------------------- */
/* Desables EPROM by reseting active low E/CE pin
 */
static void
disable_eprom( int device )
{
    if( progPinCE ) {
        /* CE and Prog are the same pin. EPRG_LATCH_CE is overridden by
         * EPRG_EPROM_PROG so we use that.
         */
        set_control( device, EPRG_EPROM_PROG );
    } else {
        set_control( device, EPRG_LATCH_CE );
    }
}

/* ------------------------------------------------------------------------- */

static void
enable_output( int device )
{
    reset_control( device, EPRG_EPROM_OE );
    /* Wait 30 ns for data to be available tGLQV */
    micro_sleep( 30 );
}

/* ------------------------------------------------------------------------- */

static void
disable_output( int device )
{
    set_control( device, EPRG_EPROM_OE );
}

/* ------------------------------------------------------------------------- */
/* Assumes device has already been selected with E/CE active low
 */
static unsigned char
read_location( int device )
{
    unsigned char byte = 0;
    
    micro_sleep(2);

    /* read byte */
    byte = read_data( device );


    return byte;
}

/* ------------------------------------------------------------------------- */

static void 
force_realtime()
{
    struct sched_param params;

    if( sched_getparam( 0, &params ) < 0 )
    {
        perror("sched_getparam");
        exit(1);
    }

    params.sched_priority = sched_get_priority_max( SCHED_FIFO );

    if( sched_setscheduler( 0, SCHED_FIFO, &params ) < 0 )
    {
        perror("sched_setscheduler");
        exit(1);
    }
}

/* ------------------------------------------------------------------------- */

static int
initialise_programmer( Device *dtype, int device )
{
    if( verbose ) printf( "Initialise programmer" );
    vindent += 2;

    /* Set all ACTIVE_LOW control pins to HIGH
     * If CE is also the Programming Pin, then switch the programmer over to:
     *   CE->NA,  PRG->P20, A14->P27 (EPRG_LATCH_CEPRG = HIGH)
     * If CE is NOT the Programming Pin, then swith the programmer over to:
     *   CE->P20, PRG->P27, A14->NA  (EPRG_LATCH_CEPRG = LOW)
     */
    set_control( device,   EPRG_EPROM_PROG     |
                           EPRG_EPROM_OE       |
                           EPRG_LATCH_CE       |
                           EPRG_LATCH_VPP_PROG |
                           EPRG_LATCH_VCC_625  |
                           (((dtype->progpin == CE) ? EPRG_LATCH_CEPRG : 0))
                        );

    if( verbose ) printf( "." );

    /* Set all ACTIVE_HIGH control pins to LOW */
    reset_control( device, EPRG_ADDR_CLOCK        |
                           EPRG_REG_LATCH         | 
                           EPRG_LATCH_CLOCK_RESET |
                           (((dtype->progpin == CE) ? 0 : EPRG_LATCH_CEPRG))
                    );

    /* Wait for everything to stabalise */
    micro_sleep(100);


    if( verbose ) printf( "." );

    /* Reset the Address Counter */
    reset_counter( device );

    if( verbose ) printf( "." );

    vindent -= 2;
    if( verbose) printf( " done\n" );

    return 0;
}

/* ------------------------------------------------------------------------- */
/* Turns off program mode
 * - Removes VPP, VCC = 5v, PROG_PULSE High
 *   Output disabled (OE HIGH), EPROM Disabled (CE High)
 */
static void
program_off( Device *dtype, int device )
{
    set_control( device, EPRG_EPROM_PROG     |
                         EPRG_EPROM_OE       |
                         EPRG_LATCH_CE       |
                         EPRG_LATCH_VPP_PROG |
                         EPRG_LATCH_VCC_625  
                       );
}

/* ------------------------------------------------------------------------- */
/* Turns on program mode
 * - Enables VPP, VCC = 6.25v if required, PROG_PULSE High
 *   Output disabled (OE HIGH), EPROM Disabled (CE High)
 */
static void
program_on( Device *dtype, int device )
{
    /* Disable OE, put PROG line HIGH 
     * IF PROG is also the CE pin (eg 27C256) then CE will thus be high
     * so for consistency, CE is also force HIGH so Eprom will always be
     * disabled on exit from this function.
     */
    set_control  ( device, EPRG_EPROM_OE   | 
                           EPRG_EPROM_PROG |   
                           EPRG_LATCH_CE   
                         );

    /* Turn on VPP and raise VCC to 6.25v if required */
    reset_control( device, EPRG_LATCH_VPP_PROG |
                           (((dtype->vcc / 1) == 6) ? EPRG_LATCH_VCC_625 : 0 ));

    if( verbose ) printf( "Enable VPP = 12.75\n%s", 
                          ((dtype->vcc / 1) == 6) ? "Enable VCC = 6.25\n":"" );

    micro_sleep(10);
}

/* ------------------------------------------------------------------------- */

static void
safe( Device *dtype, int device )
{
    /* Put programmer in safe mode */
    program_off( dtype, device );
}

/* ------------------------------------------------------------------------- */

static void
display_progress( int line, int col, int ix, int max )
{
    if( verbose > 1 ) return;

    printf( "[%d;%df", line, col );
    printf( "%5.02fK - %3d%%\n", ix/1024.00, (ix*100) / max );
    fflush( stdout );
}

/* ------------------------------------------------------------------------- */

static void
set_address( Device *dtype, int device, int address )
{
    int i = 0;
    for( i = 0; i < address; i++ ) advance_counter( device );
}

/* ------------------------------------------------------------------------- */

int
write_eprom_fast( Device *dtype, int device, char *filename, int offset )
{
    int fd      = -1;
    int  i      = 0;
    int  j      = 0;
    int  k      = 0;
    int  fail   = 0;
    int  x      = 0;
    long pulse  = 0;
    unsigned char buffer[65537];
    unsigned char c = 0;
    int  row    = 0;

    printf( "Writing device %s\n\n", dtype->device );
    row = 11;

    /* Open data to be writtng */
    if( (fd = open( filename, O_RDONLY, 0777 )) < 0 )
    {
        fprintf( stderr, "Cannot open file %s for reading\n", filename );
        return -1;
    }

    memset( (void*)buffer, 0xff, sizeof(buffer) );

    j = read( fd, &buffer, sizeof(buffer) );
    close(fd);

    if( j  < (dtype->size - offset) ) {
        printf( "File %s contains too few bytes for size of EPROM (%d < %d). Padding with 0xff\n", filename, j, (dtype->size - offset) );
        row += 3;
    }
    if( j > (dtype->size - offset) )
    {
        printf( "File %s contains too many bytes for size of EPROM (%d > %d).\n", filename, j, (dtype->size - offset) );
        return -1;
    }

    /* TODO Advance address counter to offset */

    program_on( dtype, device );
    enable_eprom( device );       /* If CE=PROG, then this does nothing */

    printf( "[%d;0fWrite ", row );


    /* CE   = LOW
     * OE   = HIGH
     * PROG = HIGH
     */
    pulse = dtype->pulse_us;

    for( i = 0; i < j; i++ )
    {
        if( i % 10 == 0 ) display_progress( row, 7, i + 1, j ); 

        write_data( device, buffer[i] );
        micro_sleep( 2 ); /* tVHEL */

        x  = 1;
        while( x <= dtype->max_pulses )
        {
            reset_control( device, EPRG_EPROM_PROG );
            micro_sleep( pulse );
            set_control  ( device, EPRG_EPROM_PROG );

            micro_sleep( 2 ); /* tQXGL */

            fail = 5; /* Read byte several time.. all must succeed */
            for( k = fail; k > 0; k-- )
            {
                /* Verify byte written */
                enable_output( device );
                c = read_location( device ); 
                disable_output( device );
            
                if( c == buffer[i] ) fail--;

                micro_sleep( 2 );
            }

            if( !fail ) {
                micro_sleep( 10 );

                /* Apply over programm pulse */
                if( dtype->overprogram )
                {
                    reset_control( device, EPRG_EPROM_PROG );
                    micro_sleep( dtype->overprogram * x );
                    set_control  ( device, EPRG_EPROM_PROG );
                }
                break; /* Success, so break out of pulse loop */
            }
            x++;
        }

        if( x > dtype->max_pulses ) break; /* Failed to prg location */

        advance_counter( device );
    }
    display_progress( row, 7, i, j );

    program_off( dtype, device ); /* Also disabled eprom */

    if( x > dtype->max_pulses ) {
        printf( "[%d;0fProgram location 0x%04x failed\n", row + 2, i + offset );
        fail = 1;
    } else {

        /* Reset the Address Counter */
        reset_counter( device );

        /* TODO Advance address counter to offset */

        /* Verify */
        enable_eprom( device );

        write_data( device, 0x00 );

        set_read_mode( device );

        micro_sleep( 25000 ); 

        printf( "[%d;0fVerify ", row );

        for( i = 0; i < dtype->size; i++ ) 
        {
            if( i % 10 == 0 ) display_progress( row, 8, i + 1, dtype->size );

            enable_output( device );
            enable_output( device );
            enable_output( device );

            c = read_location( device ); 

            if( c != buffer[i] )
            {
                printf( "[%d;0fLocation 0x%04x failed to verify\n", row + 2, i + offset );
                fail = 1;
            }

            disable_output( device );

            advance_counter( device );
        }
        /* Display final progress (ie 100% etc) */
        display_progress( row, 8, i, dtype->size );

        disable_eprom ( device );

        safe( dtype, device );
    }

    printf( "[%d;0f", row + 3 );

    if( fail ) {
        printf( "Programming failed\n" );
    } else {
        printf( "%d bytes written to EPROM from location 0x%04x\n", i, offset);
    }

    return 0;
}
/* ------------------------------------------------------------------------- */

int
write_eprom( Device *dtype, int device, char *filename, int offset )
{
    int fd      = -1;
    int  i      = 0;
    int  j      = 0;
    int  fail   = 0;
    int  x      = 0;
    long pulse  = 0;
    unsigned char buffer[65537];
    unsigned char c = 0;
    int  row    = 0;

    printf( "Writing device %s\n\n", dtype->device );

    row = 13;

    /* Open data to be writtng */
    if( (fd = open( filename, O_RDONLY, 0777 )) < 0 )
    {
        fprintf( stderr, "Cannot open file %s for reading\n", filename );
        return -1;
    }

    memset( (void*)buffer, 0xff, sizeof(buffer) );

    j = read( fd, &buffer, sizeof(buffer) );
    close(fd);

    if( j  < (dtype->size - offset) ) {
        printf( "File %s contains too few bytes for size of EPROM (%d < %d). Padding with 0xff\n", filename, j, (dtype->size - offset) );
        row += 3;
    }
    if( j > (dtype->size - offset) )
    {
        printf( "File %s contains too many bytes for size of EPROM (%d > %d).\n", filename, j, (dtype->size - offset) );
        return -1;
    }

    /* TODO Advance address counter to offset */

    program_on( dtype, device );
    enable_eprom( device );       /* If CE=PROG, then this does nothing */

    printf( "[%d;0fWrite ", row );


    /* CE   = LOW
     * OE   = HIGH
     * PROG = HIGH
     */
    pulse = dtype->pulse_us;

    /* for( i = 0; i < j; i++ ) */
    for( i = 0; i < 256 && x < 26; i++ ) 
    {
        if( i % 10 == 0 ) display_progress( row, 7, i + 1, j );

        write_data( device, buffer[i] );
        usleep( 2 ); /* tVHEL */

        x  = 0;
        while( x < 26 )
        {
            reset_control( device, EPRG_EPROM_PROG );
            micro_sleep( pulse );
            set_control  ( device, EPRG_EPROM_PROG );

            usleep( 2 ); /* tQXGL */
            /* Verify byte written */

            enable_output( device );
            c = read_location( device ); 
            disable_output( device );

            if( c == buffer[i] ) break;
            x++;
        }

        advance_counter( device );
    }
    display_progress( row, 7, i, j );

    program_off( dtype, device ); /* Also disabled eprom */

    if( x == 26 ) {
        printf( "[%d;0fProgram location 0x%04x failed\n", row + 2, i + offset - 1 );
        fail = 1;
    } else {
        /* Reset the Address Counter */
        reset_counter( device );

        /* TODO Advance address counter to offset */

        /* Verify */
        enable_eprom( device );

        write_data( device, 0x00 );

        set_read_mode( device );

        printf( "[%d;0fVerify ", row );

        /* Assert ACTIVE_LOW OE line
         */
        enable_output( device );

        for( i = 0; i < dtype->size; i++ ) 
        {
            if( i % 10 == 0 ) display_progress( row, 8, i + 1, dtype->size );

            c = read_location( device ); 
            if( c != buffer[i] )
            {
                printf( "[%d;0fLocation 0x%04x failed to verify\n", row + 2, i + offset );
                fail = 1;
            }

            advance_counter( device );
        }
        /* Display final progress (ie 100% etc) */
        display_progress( row, 8, i, dtype->size );

        disable_output( device );
        disable_eprom ( device );

        safe( dtype, device );
    }

    printf( "[%d;0f", row + 3 );

    if( fail ) {
        printf( "Programming failed\n" );
    } else {
        printf( "%d bytes written to EPROM from location 0x%4x\n", i, offset);
    }

    return 0;
}

/* ------------------------------------------------------------------------- */
/* Reads EPROM to file (if supplied).
 * Returns TRUE if EPROM is blank (all locations 0xff)
 */
int 
read_eprom( Device *dtype, int device, char *filename )
{
    int fd     = -1;
    int  i     = 0;
    unsigned char c = 0;
    unsigned char buffer[65536];
    int row = 0;
    int blank = 1;

    printf( "Reading device %s\n\n", dtype->device );
    row = 11;

    program_off( dtype, device );

    enable_eprom( device );

    write_data( device, 0x00 );

    set_read_mode( device );

    printf( "[%d;0fRead ", row );


    for( i = 0; i < dtype->size; i++ ) 
    {
        if( i % 10 == 0 ) display_progress( row, 6, i + 1, dtype->size );

        /* Assert ACTIVE_LOW OE line
         */
        enable_output( device );
        enable_output( device );
        enable_output( device );

        c = read_location( device ); 
        buffer[i] = c;
        if( c != 0xff ) blank = 0;

        disable_output( device );
        disable_output( device );
        disable_output( device );
        disable_output( device );

        advance_counter( device );
    }
    /* Display final progress (ie 100% etc) */
    display_progress( row, 6, i, dtype->size );

    disable_eprom( device );

    safe( dtype, device );

    if( filename ) {
        if( (fd = open( filename, O_RDWR | O_CREAT | O_TRUNC, 0777 )) < 0 )
        {
            fprintf( stderr, "Cannot open file %s for writing\n", filename );
        } else {
            printf( "Writing EPROM contents to %s (%d bytes)\n", filename, i );
            write( fd, buffer, i );
            close(fd);
        }
    }

    return blank;
}

/* ------------------------------------------------------------------------- */


static void 
runtest( Device *dtype, int device )
{
    int i;
    struct timeval before;
    struct timeval now;
    long diff;

    printf( "\n" );
    
    gettimeofday(&before,NULL);
    for (i=0;i<10;i++)
    {
        micro_sleep(10000);
    }
    gettimeofday(&now,NULL);
    diff=((now.tv_sec-before.tv_sec)*1000000)+(now.tv_usec-before.tv_usec);
    printf("100ms (100000us) took %luus (%+ld)\n\n",diff, diff-100000);

    gettimeofday(&before,NULL);
    micro_sleep(dtype->pulse_us);
    gettimeofday(&now,NULL);
    diff=((now.tv_sec-before.tv_sec)*1000000)+(now.tv_usec-before.tv_usec);
    printf("%ldus took %luus (%+ld)\n\n",dtype->pulse_us, diff, diff - dtype->pulse_us);

    if( dtype->overprogram )
    {
        gettimeofday(&before,NULL);
        micro_sleep(dtype->overprogram);
        gettimeofday(&now,NULL);
        diff=((now.tv_sec-before.tv_sec)*1000000)+(now.tv_usec-before.tv_usec);
        printf("%ldus took %luus (%+ld)\n\n",dtype->overprogram, diff, diff - dtype->overprogram);
    }

    prompt( "Sleep for 1 second...." );
    printf( "Sleeping\n" );
    micro_sleep(1000000);
    printf( "Sleep done\n" );

    reset_counter( device );
    set_address( dtype, device, 0x2aaa );
    prompt( "Goto address 010101010101010" );

    reset_counter( device );
    set_address( dtype, device, 0x5555 );
    prompt( "Goto address 101010101010101" );

    reset_counter( device );
    program_on( dtype, device );
    prompt( "Check Program voltages are on" );

    disable_output( device );
    prompt( "OE HIGH" );

    enable_output( device );
    prompt( "OE Low" );

    prompt( "CE HIGH" );
    disable_eprom( device );

    if( dtype->progpin != CE )
    {
        enable_eprom( device );
        prompt( "CE low" );
    }

    reset_control( device, EPRG_EPROM_PROG );
    prompt( "Program Pin LOW" );

    set_control  ( device, EPRG_EPROM_PROG );
    prompt( "Program Pin HIGH" );

    program_off( dtype, device );
    prompt( "Check Program voltages are off" );
}

/* ------------------------------------------------------------------------- */

int
main( int argc, char **argv )
{
    int     option_index           = 0;
    char    optstr[]               = "+vrbwqtDVRp:f:i:";
    int     blank                  = 0;
    char    pport [STRING_MAX+1]   = {0};
    char    filename[STRING_MAX+1] = {0};
    int     deviceID               = -1;
    int     deviceFD               = -1;
    int     realtime               = 0;
    int     testmode               = 0;
    long    offset                 = 0;
    Device *deviceType             = NULL;
    char    c                      = 0;
    char   *ep                     = NULL;

    printf( "[H[2J[H[0m" );

    printf( WELCOME );

    if( argc <= 1 ) {
        usage();
        exit(0);
    }


    /* read arguments here
    */
    while( (c = getopt_long( argc, argv, optstr, 
                    long_options, &option_index )) > (-1) ) {
        switch( c ) {
            case 'v': /* Verbose mode */
                verbose++;
                break;

            case  'i': /* Device ID */
                deviceID = strtol( optarg, (char**)&ep, 10 );
                if( *ep != '\0' ) {
                    fprintf( stderr, "Invalid ServerID\n" );
                    exit(1);
                }
                break;

            case 'p': /* set pport */
                strncpy( pport, optarg, STRING_MAX );
                break;

            case 'D': /* Display known devices */
                display_devices();
                exit(0);
                break;

            case 'r': /* Read mode */
                progMode = MODE_READ;
                break;

            case 'b': /* Check eprom is blank */
                progMode = MODE_BLANK;
                break;

            case 'w': /* Standard write mode */
                progMode = MODE_STD_WRITE;
                break;

            case 'q': /* quick write mode */
                progMode = MODE_QUICK_WRITE;
                break;

            case 'V': /* verify mode */
                progMode = MODE_VERIFY;
                break;

            case 'f': /* Supply Filename */
                strncpy( filename, optarg, STRING_MAX );
                break;

            case 't': /* Test Mode */
                testmode = 1;
                break;

            case 'R': 
                realtime = 1;
                break;

            case 'h':
                usage();
                exit(0);
                break;

            case ':':
                printf( "Missing Argument\n" );
                exit(1);
                break;

            case '?':
                printf( "Invalid Option\n" );
                break;
        }
    }

    argc -= (optind - 1);

    if( !pport[0] ) {
        fprintf( stderr, "no parallel port supplied\n" );
        exit(1);
    }

    if( deviceID < 0 ) {
        fprintf( stderr, "no device ID supplied\n" );
        exit(1);
    }

    deviceType = get_device_parameters( deviceID );
    if( !deviceType ) {
        fprintf( stderr, "Unknown device ID: %d\n", deviceID );
        exit(1);
    }


    deviceFD = open_paraport( pport );
    if( deviceFD < 0 ) exit(1);

    /* Initialise on-board control circuit and reset address counter
     */
    initialise_programmer( deviceType, deviceFD ); 


    /* Calibrate micro_sleep delay function */
    calibrate_delay_loop( deviceType, deviceFD );

    printf( "Selected device %s [%d] - %s\n", deviceType->device, deviceType->size, deviceType->manufacturer );

    progPinCE = (deviceType->progpin == CE) ? 1 : 0;;

    if( testmode ) {
        runtest( deviceType, deviceFD );
        safe( deviceType, deviceFD );
        return 0;
    }

    if( realtime ) force_realtime();

    printf( "Entering %s mode\n", ModeDescription[progMode] );

    prompt( "\nInsert EPROM into socket" );
    printf( "[9;0f%-79s[9;0f", "" );

    /* printf( "[J[6;1H[H[0m" ); */

    switch( progMode )
    {
        case MODE_BLANK: 
            printf( "Checking Eprom is Blank: " );
        case MODE_READ: 
            blank = read_eprom( deviceType, deviceFD, filename);
            if( progMode == MODE_BLANK )
            {
                if( !blank ) printf( "\n*** EPROM IS NOT BLANK ***\n" );
                if(  blank ) printf( "\nEPROM is blank\n" );
            }
            prompt( "Remove EPROM" );
            break;

        case MODE_STD_WRITE:   /* Check we have a filename */
        case MODE_QUICK_WRITE: /* Check we have a filename */
        case MODE_VERIFY:      /* Check we have a filename */
            if( !filename[0] ) {
                fprintf( stderr, "No filename supplied\n" );
                exit(1);
            }
            write_eprom_fast( deviceType, deviceFD, filename, offset );
            prompt( "Remove EPROM" );
            break;

        default:
            printf( "Huh? Unknown programmer mode\n" );
            break;
    }

    printf( "\n%s\n\n", FINISHED );

    return(0);
}

/* ------------------------------------------------------------------------- */
/* end */
